/*
 * soc_cidlo.c
 *
 * Created: 25. 1. 2020 18:15:06
 * Author : grega
 */ 

/*
	FUSE SETTINGS:
		LOW FUSE:	0x62
		HIGH FUSE:	0xD9

	voltage		ADC result
	2.4V		152
	2.3			148
	2.25V		145
*/


#define F_CPU 1000000

#define SHTC3_ADDR 0x70				//I2C address of SHTC3
#define SHTC3_SLEEP_MSB 0xB0		//high byte of sleep command
#define SHTC3_SLEEP_LSB 0x98		//low byte of sleep command
#define SHTC3_WAKEUP_MSB 0x35		//high byte of wake up command
#define SHTC3_WAKEUP_LSB 0x17		//low byte of wake up command
#define SHTC3_MEASURE_MSB_H_VALID 0x7C		//high byte of measurement command	|	temperature first, normal mode, clock stretching enabled
#define SHTC3_MEASURE_LSB_H_VALID 0xA2		//low byte of measurement command
#define SHTC3_MEASURE_MSB_T_VALID 0x5C		//high byte of measurement command	|	humidity first, normal mode, clock stretching enabled
#define SHTC3_MEASURE_LSB_T_VALID 0x24		//low byte of measurement command

#define RF_H_TENS 0
#define RF_H_ONES 1
#define RF_T_TENS 2
#define RF_T_ONES 3
#define RF_T_DECIMAL 4
#define RF_STATUS 5
#define RF_STATUS_MINUS 0
#define RF_STATUS_LOW_BATTERY 1

#define RF_NEW_DATA 0x0C
#define RF_START 0x0A
#define RF_CONTINUE 0x0D
#define RF_STOP 0x0F

#define DC_DC_pin PORTB2
#define RF_transmittion_port PORTB
#define RF_transmittion_pin PORTB0
#define RF_PORT PORTA

#define DC_DC_turn_on() PORTB |= 1<<DC_DC_pin					//power on DC/DC boost 12V converter
#define DC_DC_turn_off() PORTB &= ~(1<<DC_DC_pin)				//shut down DC/DC boost 12V converter
#define RF_transmittion_enable() RF_transmittion_port |= 1<<RF_transmittion_pin		//pin output, set to zero
#define RF_transmittion_disable() RF_transmittion_port &= ~(1<<RF_transmittion_pin)	//pin input - high impedance

#define RF_TRANSMITTING_TIME 250
#define BATTERY_VOLTAGE_MINIMUM 148

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include "USI_TWI_Master.c"

volatile unsigned char TWI_buffer[8];
volatile unsigned char RF_buffer[6];
volatile float SHTC3_Temperature;
volatile float SHTC3_Humidity;
volatile char WDTcounter = 0;

void SHTC3wakeUp(void);
void SHTC3sleep(void);
void SHTC3measure(void);
void RFsend(char RFsend_data);
void send_measured_values(void);

int main(void)
{
	WDTCSR = (1<<WDIE) | ((1<<WDP3) | (1<<WDP0));	//watchdog timer interrupt enable and prescaler set for timeout every 8s
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    DDRA = 0x0F;			//port A input for voltage measurement and I2C, output for RF
	PORTA = 0x00;
	DDRB = (1<<DC_DC_pin) | (1<<RF_transmittion_pin);	//output for DC/DC converter, and RF enable pin
	DC_DC_turn_off();
	RF_transmittion_disable();
	ADMUX = 0b10000111;					//ADC - internal 1.1V reference, ADC7 (PA7) input
	ADCSRA |= 1<<ADPS2;					//ADC prescaler set to 16 (62.5kHz)
	ADCSRB |= 1<<ADLAR;					//left adjust ADC result
	DIDR0 |= 1<<ADC7D;					//digital input disable
	USI_TWI_Master_Initialise();
	_delay_ms(100);
	SHTC3sleep();
	_delay_ms(2000);
	send_measured_values();
	
	cli();
	sleep_enable();
	sei();
	sleep_cpu();
	sleep_disable();
	
    while(1){
		asm("nop");
    }
}

ISR(WATCHDOG_vect){
	if(WDTcounter > 8){
		WDTcounter = 0;
		send_measured_values();
	}
	else{
		WDTcounter++;
	}
	WDTCSR |= 1<<WDIE;
	cli();
	sleep_enable();
	sei();
	sleep_cpu();
	sleep_disable();
}

void SHTC3wakeUp(void){
	TWI_buffer[0] = (SHTC3_ADDR << TWI_ADR_BITS) | (FALSE << TWI_READ_BIT);
	TWI_buffer[1] = SHTC3_WAKEUP_MSB;
	TWI_buffer[2] = SHTC3_WAKEUP_LSB;
	USI_TWI_Start_Transceiver_With_Data(TWI_buffer, 3);
	_delay_ms(1);
}

void SHTC3sleep(void){
	TWI_buffer[0] = (SHTC3_ADDR << TWI_ADR_BITS) | (FALSE << TWI_READ_BIT);
	TWI_buffer[1] = SHTC3_SLEEP_MSB;
	TWI_buffer[2] = SHTC3_SLEEP_LSB;
	USI_TWI_Start_Transceiver_With_Data(TWI_buffer, 3);
	_delay_ms(2);
}

void SHTC3measure(void){
	unsigned int SHTC3measureData;
	SHTC3wakeUp();
	TWI_buffer[0] = (SHTC3_ADDR << TWI_ADR_BITS) | (FALSE << TWI_READ_BIT);
	TWI_buffer[1] = SHTC3_MEASURE_MSB_H_VALID;
	TWI_buffer[2] = SHTC3_MEASURE_LSB_H_VALID;
	USI_TWI_Start_Transceiver_With_Data(TWI_buffer, 3);
	TWI_buffer[0] = (SHTC3_ADDR <<TWI_ADR_BITS) | (TRUE << TWI_READ_BIT);
	USI_TWI_Start_Transceiver_With_Data(TWI_buffer, 7);
	SHTC3sleep();
	SHTC3measureData = TWI_buffer[4]<<8;
	SHTC3measureData |= TWI_buffer[5];
	SHTC3_Humidity = (100 * (float)SHTC3measureData) / 65536;
	
	SHTC3wakeUp();
	TWI_buffer[0] = (SHTC3_ADDR << TWI_ADR_BITS) | (FALSE << TWI_READ_BIT);
	TWI_buffer[1] = SHTC3_MEASURE_MSB_T_VALID;
	TWI_buffer[2] = SHTC3_MEASURE_LSB_T_VALID;
	USI_TWI_Start_Transceiver_With_Data(TWI_buffer, 3);
	TWI_buffer[0] = (SHTC3_ADDR <<TWI_ADR_BITS) | (TRUE << TWI_READ_BIT);
	USI_TWI_Start_Transceiver_With_Data(TWI_buffer, 7);
	SHTC3sleep();
	SHTC3measureData = TWI_buffer[4]<<8;
	SHTC3measureData |= TWI_buffer[5];
	SHTC3_Temperature = (175 * (float)SHTC3measureData) / 65536;
	SHTC3_Temperature -= 45;
}

void RFsend(char RFsend_data){
	RF_PORT |= ~RFsend_data & 0x0F;
	RF_PORT &= ~RFsend_data | 0xF0;
	RF_transmittion_enable();
	_delay_ms(RF_TRANSMITTING_TIME);
	RF_transmittion_disable();
	_delay_ms(100);
}

void send_measured_values(void){
	float temperature_avg, humidity_avg;
	char batteryVoltage;
	temperature_avg = 0;
	humidity_avg = 0;
	RF_buffer[RF_STATUS] = 0x00;
	
	ADCSRA |= (1<<ADEN) | (1<<ADSC);	//turn on ADC and start conversion
	while(!(ADCSRA & (1<<ADIF)));		//wait until A/D conversion is finished
	batteryVoltage = ADCH;				//read ADC result
	ADCSRA |= 1<<ADIF;
	ADCSRA &= ~(1<<ADEN);
	
	for(char i = 10; i > 0; i--){
		SHTC3measure();
		temperature_avg += SHTC3_Temperature;
		humidity_avg += SHTC3_Humidity;
	}
	temperature_avg /= 10;
	humidity_avg /= 10;
	RF_buffer[RF_STATUS] = 0;
	RF_buffer[RF_H_TENS] = (int)humidity_avg / 10;
	RF_buffer[RF_H_ONES] = (int)humidity_avg % 10;
	if(temperature_avg < 0){
		RF_buffer[RF_STATUS] |= 1<<RF_STATUS_MINUS;
		temperature_avg *= -1;
	}
	RF_buffer[RF_T_TENS] = temperature_avg / 10;
	RF_buffer[RF_T_ONES] = (int)temperature_avg % 10;
	RF_buffer[RF_T_DECIMAL] = (int)(temperature_avg * 100.0) % 10;
	if(batteryVoltage < BATTERY_VOLTAGE_MINIMUM){
		RF_buffer[RF_STATUS] |= 1<<RF_STATUS_LOW_BATTERY;
	}
	
	DC_DC_turn_on();
	_delay_ms(500);
	RFsend(RF_NEW_DATA);
	RFsend(RF_NEW_DATA);
	for(char i = 3; i > 0; i--){
		RFsend(RF_START);
		RFsend(RF_START);
		RFsend(RF_buffer[RF_H_TENS]);
		RFsend(RF_buffer[RF_H_TENS]);
		RFsend(RF_CONTINUE);
		RFsend(RF_CONTINUE);
		RFsend(RF_buffer[RF_H_ONES]);
		RFsend(RF_buffer[RF_H_ONES]);
		RFsend(RF_CONTINUE);
		RFsend(RF_CONTINUE);
		RFsend(RF_buffer[RF_T_TENS]);
		RFsend(RF_buffer[RF_T_TENS]);
		RFsend(RF_CONTINUE);
		RFsend(RF_CONTINUE);
		RFsend(RF_buffer[RF_T_ONES]);
		RFsend(RF_buffer[RF_T_ONES]);
		RFsend(RF_CONTINUE);
		RFsend(RF_CONTINUE);
		RFsend(RF_buffer[RF_T_DECIMAL]);
		RFsend(RF_buffer[RF_T_DECIMAL]);
		RFsend(RF_CONTINUE);
		RFsend(RF_CONTINUE);
		RFsend(RF_buffer[RF_STATUS]);
		RFsend(RF_buffer[RF_STATUS]);
		RFsend(RF_STOP);
		RFsend(RF_STOP);
		_delay_ms(100);
	}
	DC_DC_turn_off();
}